@@ -61,10 +61,9 @@
  * __END_OF_JASPER_LICENSE__
  */
 
-/*
- * Math-Related Code
- *
- * $Id$
+/*!
+ * @file jas_math.h
+ * @brief Math-Related Code
  */
 
 #ifndef	JAS_MATH_H
@@ -77,10 +76,10 @@
 /* The configuration header file should be included first. */
 #include <jasper/jas_config.h>
 
+#include <jasper/jas_compiler.h>
 #include <jasper/jas_types.h>
 
 #include <assert.h>
-#include <stdio.h>
 #include <string.h>
 #include <stdint.h>
 #include <limits.h>
@@ -122,7 +121,7 @@ extern "C" {
 *
 \******************************************************************************/
 
-#ifdef __GNUC__
+#if defined(__clang__) || (defined(__GNUC__) && __GNUC__ > 6)
 /* suppress clang warning "shifting a negative signed value is
    undefined" in the assertions below */
 #pragma GCC diagnostic push
@@ -130,62 +129,84 @@ extern "C" {
 #endif
 
 JAS_ATTRIBUTE_CONST
-JAS_ATTRIBUTE_DISABLE_USAN
-inline static int jas_int_asr(int x, int n)
+JAS_ATTRIBUTE_DISABLE_UBSAN
+inline static int jas_int_asr(int x, unsigned n)
 {
 	// Ensure that the shift of a negative value appears to behave as a
 	// signed arithmetic shift.
 	assert(((-1) >> 1) == -1);
-	assert(n >= 0);
-	// The behavior is undefined when x is negative. */
+	// The behavior is undefined when x is negative.
 	// We tacitly assume the behavior is equivalent to a signed
 	// arithmetic right shift.
 	return x >> n;
 }
 
 JAS_ATTRIBUTE_CONST
-JAS_ATTRIBUTE_DISABLE_USAN
-inline static int jas_int_asl(int x, int n)
+JAS_ATTRIBUTE_DISABLE_UBSAN
+inline static int jas_int_asl(int x, unsigned n)
 {
 	// Ensure that the shift of a negative value appears to behave as a
 	// signed arithmetic shift.
 	assert(((-1) << 1) == -2);
-	assert(n >= 0);
-	// The behavior is undefined when x is negative. */
+	// The behavior is undefined when x is negative.
 	// We tacitly assume the behavior is equivalent to a signed
 	// arithmetic left shift.
 	return x << n;
 }
 
 JAS_ATTRIBUTE_CONST
-JAS_ATTRIBUTE_DISABLE_USAN
-inline static int jas_fast32_asr(int_fast32_t x, int n)
+JAS_ATTRIBUTE_DISABLE_UBSAN
+inline static int_least32_t jas_least32_asr(int_least32_t x, unsigned n)
 {
 	// Ensure that the shift of a negative value appears to behave as a
 	// signed arithmetic shift.
-	assert(((JAS_CAST(int_fast32_t, -1)) >> 1) == JAS_CAST(int_fast32_t, -1));
-	assert(n >= 0);
-	// The behavior is undefined when x is negative. */
+	assert(((JAS_CAST(int_least32_t, -1)) >> 1) == JAS_CAST(int_least32_t, -1));
+	// The behavior is undefined when x is negative.
 	// We tacitly assume the behavior is equivalent to a signed
 	// arithmetic right shift.
 	return x >> n;
 }
 
 JAS_ATTRIBUTE_CONST
-JAS_ATTRIBUTE_DISABLE_USAN
-inline static int jas_fast32_asl(int_fast32_t x, int n)
+JAS_ATTRIBUTE_DISABLE_UBSAN
+inline static int_least32_t jas_least32_asl(int_least32_t x, unsigned n)
 {
 	// Ensure that the shift of a negative value appears to behave as a
 	// signed arithmetic shift.
-	assert(((JAS_CAST(int_fast32_t, -1)) << 1) == JAS_CAST(int_fast32_t, -2));
-	assert(n >= 0);
-	// The behavior is undefined when x is negative. */
+	assert(((JAS_CAST(int_least32_t, -1)) << 1) == JAS_CAST(int_least32_t, -2));
+	// The behavior is undefined when x is negative.
 	// We tacitly assume the behavior is equivalent to a signed
 	// arithmetic left shift.
 	return x << n;
 }
 
-#ifdef __GNUC__
+JAS_ATTRIBUTE_CONST
+JAS_ATTRIBUTE_DISABLE_UBSAN
+inline static int_fast32_t jas_fast32_asr(int_fast32_t x, unsigned n)
+{
+	// Ensure that the shift of a negative value appears to behave as a
+	// signed arithmetic shift.
+	assert(((JAS_CAST(int_fast32_t, -1)) >> 1) == JAS_CAST(int_fast32_t, -1));
+	// The behavior is undefined when x is negative.
+	// We tacitly assume the behavior is equivalent to a signed
+	// arithmetic right shift.
+	return x >> n;
+}
+
+JAS_ATTRIBUTE_CONST
+JAS_ATTRIBUTE_DISABLE_UBSAN
+inline static int_fast32_t jas_fast32_asl(int_fast32_t x, unsigned n)
+{
+	// Ensure that the shift of a negative value appears to behave as a
+	// signed arithmetic shift.
+	assert(((JAS_CAST(int_fast32_t, -1)) << 1) == JAS_CAST(int_fast32_t, -2));
+	// The behavior is undefined when x is negative.
+	// We tacitly assume the behavior is equivalent to a signed
+	// arithmetic left shift.
+	return x << n;
+}
+
+#if defined(__clang__) || (defined(__GNUC__) && __GNUC__ > 6)
 #pragma GCC diagnostic pop
 #endif
 
@@ -196,6 +217,14 @@ inline static int jas_fast32_asl(int_fast32_t x, int n)
 /* Compute the product of two size_t integers with overflow checking. */
 inline static bool jas_safe_size_mul(size_t x, size_t y, size_t *result)
 {
+#if jas_has_builtin(__builtin_mul_overflow) || (defined(__GNUC__) && __GNUC__ > 5)
+	size_t result_buffer;
+	bool valid = !__builtin_mul_overflow(x, y, &result_buffer);
+	if (valid && result) {
+		*result = result_buffer;
+	}
+	return valid;
+#else
 	/* Check if overflow would occur */
 	if (x && y > SIZE_MAX / x) {
 		/* Overflow would occur. */
@@ -205,6 +234,7 @@ inline static bool jas_safe_size_mul(size_t x, size_t y, size_t *result)
 		*result = x * y;
 	}
 	return true;
+#endif
 }
 
 /* Compute the product of three size_t integers with overflow checking. */
@@ -225,6 +255,14 @@ inline static bool jas_safe_size_mul3(size_t a, size_t b, size_t c,
 /* Compute the sum of two size_t integers with overflow checking. */
 inline static bool jas_safe_size_add(size_t x, size_t y, size_t *result)
 {
+#if jas_has_builtin(__builtin_add_overflow) || (defined(__GNUC__) && __GNUC__ > 5)
+	size_t result_buffer;
+	bool valid = !__builtin_add_overflow(x, y, &result_buffer);
+	if (valid && result) {
+		*result = result_buffer;
+	}
+	return valid;
+#else
 	if (y > SIZE_MAX - x) {
 		return false;
 	}
@@ -232,11 +270,20 @@ inline static bool jas_safe_size_add(size_t x, size_t y, size_t *result)
 		*result = x + y;
 	}
 	return true;
+#endif
 }
 
 /* Compute the difference of two size_t integers with overflow checking. */
 inline static bool jas_safe_size_sub(size_t x, size_t y, size_t *result)
 {
+#if jas_has_builtin(__builtin_sub_overflow) || (defined(__GNUC__) && __GNUC__ > 5)
+	size_t result_buffer;
+	bool valid = !__builtin_sub_overflow(x, y, &result_buffer);
+	if (valid && result) {
+		*result = result_buffer;
+	}
+	return valid;
+#else
 	if (y > x) {
 		return false;
 	}
@@ -244,12 +291,21 @@ inline static bool jas_safe_size_sub(size_t x, size_t y, size_t *result)
 		*result = x - y;
 	}
 	return true;
+#endif
 }
 
 /* Compute the product of two int_fast32_t integers with overflow checking. */
 inline static bool jas_safe_intfast32_mul(int_fast32_t x, int_fast32_t y,
   int_fast32_t *result)
 {
+#if jas_has_builtin(__builtin_mul_overflow) || (defined(__GNUC__) && __GNUC__ > 5)
+	int_fast32_t result_buffer;
+	bool valid = !__builtin_mul_overflow(x, y, &result_buffer);
+	if (valid && result) {
+		*result = result_buffer;
+	}
+	return valid;
+#else
 	if (x > 0) {
 		/* x is positive */
 		if (y > 0) {
@@ -281,6 +337,7 @@ inline static bool jas_safe_intfast32_mul(int_fast32_t x, int_fast32_t y,
 		*result = x * y;
 	}
 	return true;
+#endif
 }
 
 /* Compute the product of three int_fast32_t integers with overflow checking. */
@@ -302,6 +359,14 @@ inline static bool jas_safe_intfast32_mul3(int_fast32_t a, int_fast32_t b,
 inline static bool jas_safe_intfast32_add(int_fast32_t x, int_fast32_t y,
   int_fast32_t *result)
 {
+#if jas_has_builtin(__builtin_add_overflow) || (defined(__GNUC__) && __GNUC__ > 5)
+	int_fast32_t result_buffer;
+	bool valid = !__builtin_add_overflow(x, y, &result_buffer);
+	if (valid && result) {
+		*result = result_buffer;
+	}
+	return valid;
+#else
 	if ((y > 0 && x > INT_FAST32_MAX - y) ||
 	  (y < 0 && x < INT_FAST32_MIN - y)) {
 		return false;
@@ -310,8 +375,137 @@ inline static bool jas_safe_intfast32_add(int_fast32_t x, int_fast32_t y,
 		*result = x + y;
 	}
 	return true;
+#endif
 }
 
+#if 0
+/*
+This function is potentially useful but not currently used.
+So, it is commented out.
+*/
+inline static bool jas_safe_uint_mul(unsigned x, unsigned y, unsigned *result)
+{
+	/* Check if overflow would occur */
+	if (x && y > UINT_MAX / x) {
+		/* Overflow would occur. */
+		return false;
+	}
+	if (result) {
+		*result = x * y;
+	}
+	return true;
+}
+#endif
+
+/******************************************************************************\
+* Safe integer arithmetic (i.e., with overflow checking).
+\******************************************************************************/
+
+typedef struct {
+	bool valid;
+	uint_least64_t value;
+} jas_safeui64_t;
+
+static inline
+jas_safeui64_t jas_safeui64_from_intmax(intmax_t x)
+{
+	jas_safeui64_t result;
+	if (x >= 0 && x <= UINT_LEAST64_MAX) {
+		result.valid = true;
+		result.value = JAS_CAST(uint_least64_t, x);
+	} else {
+		result.valid = false;
+		result.value = 0;
+	}
+	return result;
+}
+
+static inline
+jas_safeui64_t jas_safeui64_add(jas_safeui64_t x, jas_safeui64_t y)
+{
+	jas_safeui64_t result;
+	if (x.valid && y.valid && y.value <= UINT_LEAST64_MAX - x.value) {
+		result.valid = true;
+		result.value = x.value + y.value;
+	} else {
+		result.valid = false;
+		result.value = 0;
+	}
+	return result;
+}
+
+static inline
+jas_safeui64_t jas_safeui64_sub(jas_safeui64_t x, jas_safeui64_t y)
+{
+	jas_safeui64_t result;
+	if (x.valid && y.valid && y.value <= x.value) {
+		result.valid = true;
+		result.value = x.value - y.value;
+	} else {
+		result.valid = false;
+		result.value = 0;
+	}
+	return result;
+}
+
+static inline
+jas_safeui64_t jas_safeui64_mul(jas_safeui64_t x, jas_safeui64_t y)
+{
+	jas_safeui64_t result;
+	if (!x.valid || !y.valid || (x.value && y.value > UINT_LEAST64_MAX /
+	  x.value)) {
+		result.valid = false;
+		result.value = 0;
+	} else {
+		result.valid = true;
+		result.value = x.value * y.value;
+	}
+	return result;
+}
+
+static inline
+jas_safeui64_t jas_safeui64_div(jas_safeui64_t x, jas_safeui64_t y)
+{
+	jas_safeui64_t result;
+	if (x.valid && y.valid && y.value) {
+		result.valid = true;
+		result.value = x.value / y.value;
+	} else {
+		result.valid = false;
+		result.value = 0;
+	}
+	return result;
+}
+
+static inline
+jas_safeui64_t jas_safeui64_pow2_intmax(intmax_t x)
+{
+	jas_safeui64_t result;
+	if (x >= 0 && x < 64) {
+		result.valid = true;
+		result.value = JAS_CAST(uint_least64_t, 1) << x;
+	} else {
+		result.valid = false;
+		result.value = 0;
+	}
+	return result;
+}
+
+static inline
+int jas_safeui64_to_int(jas_safeui64_t x, int invalid_value)
+{
+	int result;
+	if (x.valid && x.value <= INT_MAX) {
+		result = JAS_CAST(int, x.value);
+	} else {
+		result = invalid_value;
+	}
+	return result;
+}
+
+/******************************************************************************\
+\******************************************************************************/
+
 #ifdef __cplusplus
 }
 #endif
